import socket
import select
import queue
from collections import defaultdict
import logging
from companion.handler import HttpRequestHandler
from companion.parser import HttpParser
from pathlib import Path

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


HOST = "localhost"
PORT = 8181
CHUNK_SIZE = 1024
END_HTTP_REQUEST = "\r\n\r\n"

STATIC_FILE_DIR = Path("C:/Users/dankr/OneDrive/Desktop/companion/testfolder")

inputs = []
outputs = []
exceptions = []

messages = defaultdict(queue.Queue)
request_handler = HttpRequestHandler(file_directory=STATIC_FILE_DIR)


def read_http_request(connection):
    found_end = False
    data = b""
    while not found_end:
        data += connection.recv(CHUNK_SIZE)
        if not data:
            break
        if data.decode("ascii")[-4:] == END_HTTP_REQUEST:
            found_end = True
    return data


def handle_read(connection):
    http_request_bytes = read_http_request(connection)
    http_request = HttpParser(http_request_bytes).parse()
    logger.info(f"Incoming Request {http_request}")
    http_response = request_handler.handle(http_request)
    messages[connection].put_nowait(http_response.bytes)


def handle_write(connection: socket.socket):
    message = messages[connection].get_nowait()
    logger.info(f"Sending Response {message}")
    connection.sendall(message)
    

def handle_exception(connection):
    logger.info("Handling exception connection")
    ...


def initialize_server_socket() -> socket.socket:
    logger.info("Creating Server socket")
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setblocking(0)
    sock.bind((HOST, PORT))
    sock.listen(5)
    return sock


def run_forever():
    server_socket = initialize_server_socket()
    inputs.append(server_socket)
    try:
        while True:
            read_list, write_list, exception_list = select.select(inputs, outputs, exceptions)
            for conn in read_list:
                if conn == server_socket:
                    new_connection, address = conn.accept()
                    new_connection.setblocking(0)
                    inputs.append(new_connection)
                    break
                handle_read(conn)
                inputs.remove(conn)
                outputs.append(conn)
            for conn in write_list:
                handle_write(conn)
                outputs.remove(conn)
                conn.close()
            for conn in exception_list:
                handle_exception(conn)
    except Exception as exc:
        logger.exception(exc)
        server_socket.close()


def cli():
    run_forever()


from companion.enums import HttpStatus


class HttpResponse(object):
    CRLF = "\r\n"

    def __init__(self, status_code: HttpStatus, *, version = "HTTP/1.0", headers=None, body=None):
        self.status_code_message = status_code.message
        self.status_code = status_code.value
        self.version = version
        self.body = body
        self.headers = headers

    @property
    def bytes(self) -> bytes:
        response = ""
        response += self._build_response_line()
        response += self.CRLF
        if self.headers:
            response += self._build_headers()
            response += self.CRLF
        response += self.CRLF
        return bytes(response, encoding="ascii") + self._build_response_body()

    
    def _build_response_line(self) -> str:
        return " ".join([self.version, str(self.status_code), self.status_code_message])

    def _build_headers(self) -> str:
        return self.CRLF.join([f"{k}: {v}" for k, v in self.headers.items()])
    
    def _build_response_body(self) -> str:
        return self.body if self.body else b""

    def __str__(self):
        return f"HttpResponse(status_code={self.status_code}, headers={self.headers}, body={self.body})"

    def __repr__(self):
        return f"HttpResponse(version={self.version}, message={self.status_code_message}, status_code={self.status_code}, headers={self.headers}, body={self.body})"


from enum import Enum, auto


class HttpMethod(Enum):
    GET = auto()
    HEAD = auto()


class HttpStatus(Enum):
    OK = (200, "OK")
    NOT_FOUND = (404, "Not Found")

    def __new__(cls, value, message):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.message = message
        return obj



from companion.enums import HttpMethod


class HttpRequestLine(object):
    def __init__(self, method: HttpMethod, target: str, version: str):
        self.method = method
        self.target = target
        self.version = version

    def __str__(self):
        return f"HttpRequestLine(method={self.method}, target={self.target}, version={self.version})"

    def __repr__(self):
        return f"HttpRequestLine(method={self.method}, target={self.target}, version={self.version})"



class HttpRequest(object):
    def __init__(self, request_line: HttpRequestLine, headers=None, body=None):
        self.request_line = request_line
        self.method = self.request_line.method
        self.headers = headers
        self.body = body

    def __str__(self):
        return f"HttpRequest(request_line={self.request_line}, headers={self.headers}, body={self.body})"

    def __repr__(self):
        return f"HttpRequest(request_line={self.request_line}, headers={self.headers}, body={self.body})"


from companion.request import HttpRequest, HttpRequestLine
from companion.enums import HttpMethod


class HttpParser(object):
    """A simple HTTP Parser that turns bytes into HttpRequests"""
    CRLF = "\r\n"

    def __init__(self, data: bytes):
        self.raw_data = data
    
    def parse(self) -> HttpRequest:
        decoded_string = self.raw_data.decode("ascii")
        lines = decoded_string.split(self.CRLF)
        request_line: HttpRequestLine = self._parse_request_line(lines[0])
        headers = self._parse_headers(lines[1:])
        return HttpRequest(request_line, headers)
    
    def _parse_request_line(self, request_line: str) -> HttpRequestLine:
        components = request_line.split()
        method = HttpMethod[components[0]]
        uri = components[1]
        version = components[2]
        return HttpRequestLine(method, uri, version)

    def _parse_headers(self, potential_headers: list[str]):
        headers = {}
        for header in potential_headers:
            if header:
                header_split = header.split(":")
                key = header_split[0]
                val = header_split[1].lstrip()
                headers[key] = val
        return headers
                




from companion.enums import HttpMethod, HttpStatus
from companion.request import HttpRequest
from companion.response import HttpResponse
from pathlib import Path


class HttpRequestHandler(object):
    def __init__(self, file_directory: Path):
        self.file_directory = file_directory

    def handle(self, http_request: HttpRequest):
        match http_request.method:
            case HttpMethod.GET:
                return self.get(http_request.request_line.target, http_request.headers)

    def get(self, request_target: str, headers: dict):
        clean_request_target = request_target.lstrip("/")
        target_file_or_directory = (self.file_directory / clean_request_target).resolve()

        if self.file_directory in target_file_or_directory.parents or target_file_or_directory == self.file_directory:
            if target_file_or_directory.exists():
                if target_file_or_directory.is_dir():
                    if (target_file_or_directory / "index.html").exists():
                        with (target_file_or_directory / "index.html").open("rb") as fp:
                            content = fp.read()
                            return HttpResponse(HttpStatus.OK, body=content)
                else:
                    with target_file_or_directory.open("rb") as fp:
                        content = fp.read()
                        return HttpResponse(HttpStatus.OK, body=content)
        return HttpResponse(HttpStatus.NOT_FOUND)

    def head(self, request_target: str, headers: dict):
        ...
